@@ -22,6 +22,13 @@ from functools import reduce
 import xlrd
 import threading
 import math
+import platform
+from matplotlib.font_manager import FontManager
+import subprocess
+import importlib
+import shutil
+import random
+from numba import njit
 
 logger = logging.getLogger(__name__)
 STR_FORMAT_DATE = '%Y-%m-%d'
@@ -208,6 +215,19 @@ def str_2_float(sth) -> (float, None):
     return ret_val
 
 
+def format_2_str(value, formator):
+    """根据 formator 将对象格式化成 str"""
+    if formator is None:
+        text = str(value)
+    elif isinstance(formator, str):
+        text = str.format(formator, value)
+    elif callable(formator):
+        text = formator(value)
+    else:
+        raise ValueError('%s: %s 无效', value, formator)
+    return text
+
+
 class TryThread(threading.Thread):
 
     def __init__(self, target, *args, **kwargs):
@@ -465,10 +485,10 @@ def pd_timedelta_2_timedelta(value):
     return dt_value
 
 
-def get_first(iterable, func):
-    for n in iterable:
-        if func(n):
-            return n
+def get_first(iterable, func, ret_func=None):
+    for val in iterable:
+        if func(val):
+            return val if ret_func is None else ret_func(val)
     return None
 
 
@@ -479,11 +499,28 @@ def get_first_idx(iterable, func):
     return None
 
 
-def get_last(iterable, func):
+@njit
+def get_first_idx_larger_than(arr: np.ndarray, k):
+    for i in range(len(arr)):
+        if arr[i] > k:
+            return i
+    return -1
+
+
+@njit
+def get_first_idx_smaller_than(arr: np.ndarray, k):
+    for i in range(len(arr)):
+        if arr[i] < k:
+            return i
+    return -1
+
+
+def get_last(iterable, comp_func, ret_func=None):
     count = len(iterable)
     for n in range(count - 1, -1, -1):
-        if func(iterable[n]):
-            return iterable[n]
+        val = iterable[n]
+        if comp_func(val):
+            return val if ret_func is None else ret_func(val)
     return None
 
 
@@ -501,6 +538,83 @@ def get_last_idx(iterable, func):
     return None
 
 
+@njit
+def get_last_idx_larger_than(arr: np.ndarray, k):
+    for i in range(len(arr) - 1, -1, -1):
+        if arr[i] > k:
+            return i
+    return -1
+
+
+@njit
+def get_last_idx_smaller_than(arr: np.ndarray, k):
+    for i in range(len(arr) - 1, -1, -1):
+        if arr[i] < k:
+            return i
+    return -1
+
+
+@njit
+def get_nth_index(arr: np.ndarray, func, count):
+    c = 0
+    for i in range(len(arr)):
+        if func(arr[i]):
+            c += 1
+            if c == count:
+                return i
+    return -1
+
+
+@njit
+def get_nth_index_right(arr: np.ndarray, func, count):
+    c = 0
+    for i in range(len(arr) - 1, -1, -1):
+        if func(arr[i]):
+            c += 1
+            if c == count:
+                return i
+    return -1
+
+
+def _test_get_idx_nb():
+    np.random.seed(0)
+    arr = np.random.rand(10 ** 7)
+    m = 0.999999
+    n = 0.9999999
+    # # Start of array benchmark
+    # % timeit next(iter(np.where(arr > m)[0]), -1)  # 43.5 ms
+    # % timeit next((idx for idx, val in enumerate(arr) if val > m), -1)  # 2.5 µs
+    # # End of array benchmark
+    # % timeit next(iter(np.where(arr > n)[0]), -1)  # 21.4 ms
+    # % timeit next((idx for idx, val in enumerate(arr) if val > n), -1)  # 39.2 ms
+    idx = get_first_idx_larger_than(arr, n)
+    assert idx == -1
+    idx = get_first_idx_smaller_than(arr, 1 - n)
+    assert idx == 3600965
+    idx = get_last_idx_larger_than(arr, n)
+    assert idx == -1
+    idx = get_last_idx_smaller_than(arr, 1 - n)
+    assert idx == 3600965
+
+    idx = get_first_idx_larger_than(arr, m)
+    assert idx == 198253
+    idx = get_first_idx_smaller_than(arr, 1 - m)
+    assert idx == 661553
+    idx = get_last_idx_larger_than(arr, m)
+    assert idx == 8361873
+    idx = get_last_idx_smaller_than(arr, 1 - m)
+    assert idx == 6590717
+
+    @njit
+    def func(val):
+        return val > m
+
+    idx = get_nth_index(arr, func, 2)
+    assert idx == 801807
+    idx = get_nth_index_right(arr, func, 2)
+    assert idx == 8142781
+
+
 def replace_none_2_str(string, replace=''):
     return replace if string is None else string
 
@@ -573,6 +687,19 @@ def get_cntr_kind_name(contract_name):
 CACHE_FOLDER_PATH_DIC = {}
 
 
+def is_pattern_type(obj):
+    """
+    判断当前对象是否是 Pattern 类
+    3.7以后 re._pattern_type 被废弃，re.Pattern开始启用。此写法为了兼容3.6. 以及 3.7 以后版本
+    :param obj:
+    :return:
+    """
+    try:
+        return isinstance(obj, re.Pattern)
+    except AttributeError:
+        return isinstance(obj, re._pattern_type)
+
+
 def get_folder_path(target_folder_name=None, create_if_not_found=True):
     """
     获得系统缓存目录路径
@@ -593,7 +720,7 @@ def get_folder_path(target_folder_name=None, create_if_not_found=True):
             dir_list = os.listdir(par_path)
             for dir_name in dir_list:
                 # print d # .strip()
-                if isinstance(target_folder_name, re._pattern_type):
+                if is_pattern_type(target_folder_name):
                     match = target_folder_name.match(dir_name)
                     if match is not None:
                         cache_folder_path_tmp = os.path.join(par_path, dir_name)
@@ -607,7 +734,7 @@ def get_folder_path(target_folder_name=None, create_if_not_found=True):
                 break
             par_path = os.path.abspath(os.path.join(par_path, os.path.pardir))
         if cache_folder_path_tmp is None:
-            if create_if_not_found and not isinstance(target_folder_name, re._pattern_type):
+            if create_if_not_found and not is_pattern_type(target_folder_name):
                 cache_folder_path_tmp = os.path.abspath(os.path.join(parent_folder_path, target_folder_name))
                 logger.debug('<%s> 创建缓存目录', cache_folder_path_tmp)
                 os.makedirs(cache_folder_path_tmp)
@@ -1427,7 +1554,6 @@ def decorator_timer(func):
 
 
 def open_file_with_system_app(file_path, asyn=True):
-    import platform
     try:
         if platform.system() == 'Windows':
             os.startfile(file_path)
@@ -1448,6 +1574,32 @@ def open_file_with_system_app(file_path, asyn=True):
         webbrowser.open(f'file:///{file_path}')
 
 
+def is_windows_os():
+    return platform.system() == 'Windows'
+
+
+def is_linux_os():
+    return platform.system() == 'Linux'
+
+
+def get_chinese_font_iter():
+    fm = FontManager()
+    mat_fonts = set(f.name for f in fm.ttflist)
+
+    output = subprocess.check_output(
+        'fc-list :lang=zh -f "%{family}\n"', shell=True)
+    output = output.decode('utf-8')
+    # print '*' * 10, '系统可用的中文字体', '*' * 10
+    # print output
+    zh_fonts = set(f.split(',', 1)[0] for f in output.split('\n'))
+    available = mat_fonts & zh_fonts
+    yield from available
+
+    # print('*' * 10, '可用的字体', '*' * 10)
+    # for num, f in enumerate(available, start=1):
+    #     print(num, ')', f)
+
+
 def get_project_root_path():
     import sys
     import os
@@ -1469,12 +1621,147 @@ def get_module_path(stg_class: type):
     module_path = stg_class.__module__
     if module_path == '__main__':
         project_root_path, work_path = get_project_root_path()
-        module_file_path = os.path.splitext(sys.argv[0])[0]
+        # windows 环境下，sys.argv[0] 中路径分割为 "/" 而系统分隔符为 r"\" 导致匹配失败，因此需要进行一次转换
+        module_file_path = os.path.splitext(sys.argv[0])[0].replace('/', os.path.sep)
         module_segment = [str(_) for _ in module_file_path[len(project_root_path):].split(os.path.sep) if _ != '']
         module_path = '.'.join(module_segment)
     return module_path
 
 
+def counting_years(date_from_str, date_to_str):
+    """
+    用于计算两个日期之间相差多少年？返回float类型
+    :param date_from_str:
+    :param date_to_str:
+    :return:
+    """
+    # date_from_str, date_to_str = '2018-05-01', '2019-05-31'
+    date_from = str_2_date(date_from_str)
+    date_to = str_2_date(date_to_str)
+
+    year_start = str_2_date(f'{date_from.year}-01-01')
+    year_end = str_2_date(f'{date_from.year}-12-31')
+    ret_val1 = ((year_end - date_from).days + 1) / ((year_end - year_start).days + 1)
+
+    year_start = str_2_date(f'{date_to.year}-01-01')
+    year_end = str_2_date(f'{date_to.year}-12-31')
+    ret_val2 = ((date_to - year_start).days + 1) / ((year_end - year_start).days + 1)
+
+    ret_val = ret_val1 + ret_val2 + (date_to.year - date_from.year - 1)
+
+    return ret_val
+
+
+def get_module_file_path(stg_class: type):
+    module_path = stg_class.__module__
+    module = importlib.import_module(module_path)
+    return module.__file__
+
+
+def _test_get_module_file_path():
+    # from ibats_common.example.tflearn.lstm3_stg import AIStg
+    file_path = get_module_file_path(DataFrame)
+    print(file_path)
+
+
+def copy_module_file_to(module_str_or_class, folder_path):
+    """将模板备份到指定目录下，不改变文件名"""
+    if isinstance(module_str_or_class, str):
+        module = importlib.import_module(module_str_or_class)
+    # elif hasattr(module_str_or_class, '__module__'):
+    #     module = importlib.import_module(module_str_or_class.__module__)
+    elif isinstance(module_str_or_class, type):
+        module = importlib.import_module(module_str_or_class.__module__)
+    elif hasattr(module_str_or_class, '__file__'):
+        module =module_str_or_class
+    else:
+        raise ValueError(f'{module_str_or_class} <{type(module_str_or_class)}> 不是有效的对象')
+
+    file_path = module.__file__
+    # _, file_name = os.path.split(file_path)
+    # if not os.path.exists(folder_path):
+    #     os.makedirs(folder_path)
+    # new_file_path = os.path.join(folder_path, file_name)
+    # shutil.copy(file_path, new_file_path)
+    # return new_file_path
+    return copy_file_to(file_path, folder_path)
+
+
+def copy_file_to(file_path, folder_path):
+    """将文件拷贝到指定目录"""
+    _, file_name = os.path.split(file_path)
+    if not os.path.exists(folder_path):
+        os.makedirs(folder_path)
+    new_file_path = os.path.join(folder_path, file_name)
+    shutil.copy(file_path, new_file_path)
+    return new_file_path
+
+
+def _test_copy_module_file_to():
+    copy_module_file_to(DataFrame, r'd:\Downloads')
+
+
+def copy_folder_to(source_folder_path, target_folder_path):
+    """将文件拷贝到指定目录，文件名不变"""
+    _, folder_name = os.path.split(source_folder_path)
+    new_folder_path = os.path.join(target_folder_path, folder_name)
+    if os.path.exists(new_folder_path):
+        shutil.rmtree(new_folder_path)
+    shutil.copytree(source_folder_path, new_folder_path)
+    return new_folder_path
+
+
+def _test_copy_folder_to():
+    folder_path = r'/home/mg/github/IBATS_Common/ibats_common/example/drl/d3qn1'
+    copy_folder_to(folder_path, r'/home/mg/Downloads')
+
+
+def sample_weighted(ll, weights, k):
+    """
+    带权重、非重复取样
+    :param ll:
+    :param weights:
+    :param k:
+    :return:
+    """
+    ll_len = len(ll)
+    if ll_len < k:
+        raise ValueError(f"len(ll)={ll_len}<{k}")
+    elif ll_len == k:
+        return random.sample(ll, k)
+    else:
+        new_ll = list(range(len(ll)))
+        new_ll_set = set(new_ll)
+        new_weights = np.array(weights)
+        new_k = k
+        result_tot = []
+        while True:
+            result = random.choices(new_ll, weights=new_weights[new_ll], k=new_k)
+            result_set = set(result)
+            result_tot.extend(result_set)
+            new_ll_set -= result_set
+            new_k = k - len(result_tot)
+            new_ll = list(new_ll_set)
+            if new_k == 0:
+                break
+
+        result_tot.sort()
+        if isinstance(ll, np.ndarray):
+            ret = ll[result_tot]
+        else:
+            ret = [ll[_] for _ in result_tot]
+        return ret
+
+
+def _test_weighted_sample():
+    ll = list(range(1, 30))
+    result = sample_weighted(ll, weights=ll, k=15)
+    print(result)
+    ll = np.eye(29)
+    result = sample_weighted(ll, weights=np.arange(1, 30), k=15)
+    print(result)
+
+
 if __name__ == "__main__":
     pass
     # logging.basicConfig(level=logging.DEBUG,
@@ -1577,3 +1864,8 @@ if __name__ == "__main__":
     #     raise Exception('some error')
     #
     # foo(1, 2, 3, 4, e=5, f=6)
+
+    # _test_get_module_file_path()
+    # _test_copy_module_file_to()
+    # _test_copy_folder_to()
+    _test_get_idx_nb()
